@@ -15,8 +15,8 @@
  */
 #include "config.h"
 #include "rspamd.h"
-#include "map.h"
-#include "filter.h"
+#include "libserver/maps/map.h"
+#include "scan_result.h"
 #include "dynamic_cfg.h"
 #include "unix-std.h"
 #include "lua/lua_common.h"
@@ -283,7 +283,7 @@ init_dynamic_config (struct rspamd_config *cfg)
 			json_config_read_cb,
 			json_config_fin_cb,
 			json_config_dtor_cb,
-			(void **)pjb)) {
+			(void **)pjb, NULL, RSPAMD_MAP_DEFAULT)) {
 		msg_err ("cannot add map for configuration %s", cfg->dynamic_conf);
 	}
 }
