@@ -98,22 +98,30 @@ struct lua_redis_userdata {
 	struct rspamd_task *task;
 	struct rspamd_symcache_item *item;
 	struct rspamd_async_session *s;
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	struct rspamd_config *cfg;
 	struct rspamd_redis_pool *pool;
 	gchar *server;
+	gchar log_tag[RSPAMD_LOG_ID_LEN + 1];
 	struct lua_redis_request_specific_userdata *specific;
 	gdouble timeout;
 	guint16 port;
 	guint16 terminated;
 };
 
+#define msg_debug_lua_redis(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_lua_redis_log_id, "lua_redis", ud->log_tag, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+INIT_LOG_MODULE(lua_redis)
+
 #define LUA_REDIS_SPECIFIC_REPLIED (1 << 0)
 /* session was finished */
 #define LUA_REDIS_SPECIFIC_FINISHED (1 << 1)
 #define LUA_REDIS_ASYNC (1 << 0)
 #define LUA_REDIS_TEXTDATA (1 << 1)
 #define LUA_REDIS_TERMINATED (1 << 2)
+#define LUA_REDIS_NO_POOL (1 << 3)
 #define IS_ASYNC(ctx) ((ctx)->flags & LUA_REDIS_ASYNC)
 
 struct lua_redis_request_specific_userdata {
@@ -124,7 +132,7 @@ struct lua_redis_request_specific_userdata {
 	struct lua_redis_userdata *c;
 	struct lua_redis_ctx *ctx;
 	struct lua_redis_request_specific_userdata *next;
-	struct event timeout;
+	ev_timer timeout_ev;
 	guint flags;
 };
 
@@ -179,14 +187,12 @@ lua_redis_dtor (struct lua_redis_ctx *ctx)
 	struct redisAsyncContext *ac;
 
 	ud = &ctx->async;
-	msg_debug ("desctructing %p", ctx);
+	msg_debug_lua_redis ("desctructing %p", ctx);
 
 	if (ud->ctx) {
 
 		LL_FOREACH_SAFE (ud->specific, cur, tmp) {
-			if (rspamd_event_pending (&cur->timeout, EV_TIMEOUT)) {
-				event_del (&cur->timeout);
-			}
+			ev_timer_stop (ud->event_loop, &cur->timeout_ev);
 
 			if (!(cur->flags & LUA_REDIS_SPECIFIC_REPLIED)) {
 				is_successful = FALSE;
@@ -200,7 +206,17 @@ lua_redis_dtor (struct lua_redis_ctx *ctx)
 		ud->terminated = 1;
 		ac = ud->ctx;
 		ud->ctx = NULL;
-		rspamd_redis_pool_release_connection (ud->pool, ac, !is_successful);
+
+		if (!is_successful) {
+			rspamd_redis_pool_release_connection (ud->pool, ac,
+					RSPAMD_REDIS_RELEASE_FATAL);
+		}
+		else {
+			rspamd_redis_pool_release_connection (ud->pool, ac,
+					(ctx->flags & LUA_REDIS_NO_POOL) ?
+					RSPAMD_REDIS_RELEASE_ENFORCE : RSPAMD_REDIS_RELEASE_DEFAULT);
+		}
+
 	}
 
 	LL_FOREACH_SAFE (ud->specific, cur, tmp) {
@@ -241,15 +257,16 @@ static void
 lua_redis_fin (void *arg)
 {
 	struct lua_redis_request_specific_userdata *sp_ud = arg;
+	struct lua_redis_userdata *ud;
 	struct lua_redis_ctx *ctx;
 
 	ctx = sp_ud->ctx;
+	ud = sp_ud->c;
 
-	if (rspamd_event_pending (&sp_ud->timeout, EV_TIMEOUT)) {
-		event_del (&sp_ud->timeout);
-	}
+	ev_timer_stop (sp_ud->ctx->async.event_loop, &sp_ud->timeout_ev);
 
-	msg_debug ("finished redis query %p from session %p", sp_ud, ctx);
+	msg_debug_lua_redis ("finished redis query %p from session %p; refcount=%d",
+			sp_ud, ctx, ctx->ref.refcount);
 	sp_ud->flags |= LUA_REDIS_SPECIFIC_FINISHED;
 
 	REDIS_RELEASE (ctx);
@@ -268,12 +285,16 @@ lua_redis_push_error (const gchar *err,
 {
 	struct lua_redis_userdata *ud = sp_ud->c;
 	struct lua_callback_state cbs;
+	lua_State *L;
 
 	if (!(sp_ud->flags & (LUA_REDIS_SPECIFIC_REPLIED|LUA_REDIS_SPECIFIC_FINISHED))) {
 		if (sp_ud->cbref != -1) {
 
 			lua_thread_pool_prepare_callback (ud->cfg->lua_thread_pool, &cbs);
+			L = cbs.L;
 
+			lua_pushcfunction (L, &rspamd_lua_traceback);
+			int err_idx = lua_gettop (L);
 			/* Push error */
 			lua_rawgeti (cbs.L, LUA_REGISTRYINDEX, sp_ud->cbref);
 
@@ -286,11 +307,11 @@ lua_redis_push_error (const gchar *err,
 				rspamd_symcache_set_cur_item (ud->task, ud->item);
 			}
 
-			if (lua_pcall (cbs.L, 2, 0, 0) != 0) {
+			if (lua_pcall (cbs.L, 2, 0, err_idx) != 0) {
 				msg_info ("call to callback failed: %s", lua_tostring (cbs.L, -1));
-				lua_pop (cbs.L, 1);
 			}
 
+			lua_settop (L, err_idx - 1);
 			lua_thread_pool_restore_callback (&cbs);
 		}
 
@@ -360,11 +381,15 @@ lua_redis_push_data (const redisReply *r, struct lua_redis_ctx *ctx,
 {
 	struct lua_redis_userdata *ud = sp_ud->c;
 	struct lua_callback_state cbs;
+	lua_State *L;
 
 	if (!(sp_ud->flags & (LUA_REDIS_SPECIFIC_REPLIED|LUA_REDIS_SPECIFIC_FINISHED))) {
 		if (sp_ud->cbref != -1) {
 			lua_thread_pool_prepare_callback (ud->cfg->lua_thread_pool, &cbs);
+			L = cbs.L;
 
+			lua_pushcfunction (L, &rspamd_lua_traceback);
+			int err_idx = lua_gettop (L);
 			/* Push error */
 			lua_rawgeti (cbs.L, LUA_REGISTRYINDEX, sp_ud->cbref);
 			/* Error is nil */
@@ -376,11 +401,11 @@ lua_redis_push_data (const redisReply *r, struct lua_redis_ctx *ctx,
 				rspamd_symcache_set_cur_item (ud->task, ud->item);
 			}
 
-			if (lua_pcall (cbs.L, 2, 0, 0) != 0) {
+			if (lua_pcall (cbs.L, 2, 0, err_idx) != 0) {
 				msg_info ("call to callback failed: %s", lua_tostring (cbs.L, -1));
-				lua_pop (cbs.L, 1);
 			}
 
+			lua_settop (L, err_idx - 1);
 			lua_thread_pool_restore_callback (&cbs);
 		}
 
@@ -422,7 +447,8 @@ lua_redis_callback (redisAsyncContext *c, gpointer r, gpointer priv)
 		return;
 	}
 
-	msg_debug ("got reply from redis %p for query %p", sp_ud->c->ctx, sp_ud);
+	msg_debug_lua_redis ("got reply from redis %p for query %p", sp_ud->c->ctx,
+			sp_ud);
 
 	REDIS_RETAIN (ctx);
 
@@ -460,7 +486,11 @@ lua_redis_callback (redisAsyncContext *c, gpointer r, gpointer priv)
 		ud->ctx = NULL;
 
 		if (ac) {
-			rspamd_redis_pool_release_connection (ud->pool, ac, FALSE);
+			msg_debug_lua_redis ("release redis connection ud=%p; ctx=%p; refcount=%d",
+					ud, ctx, ctx->ref.refcount);
+			rspamd_redis_pool_release_connection (ud->pool, ac,
+					(ctx->flags & LUA_REDIS_NO_POOL) ?
+					RSPAMD_REDIS_RELEASE_ENFORCE : RSPAMD_REDIS_RELEASE_DEFAULT);
 		}
 	}
 
@@ -556,11 +586,8 @@ lua_redis_callback_sync (redisAsyncContext *ac, gpointer r, gpointer priv)
 		return;
 	}
 
-	if (rspamd_event_pending (&sp_ud->timeout, EV_TIMEOUT)) {
-		event_del (&sp_ud->timeout);
-	}
-
-	msg_debug ("got reply from redis: %p for query %p", ac, sp_ud);
+	ev_timer_stop (ud->event_loop, &sp_ud->timeout_ev);
+	msg_debug_lua_redis ("got reply from redis: %p for query %p", ac, sp_ud);
 
 	struct lua_redis_result *result = g_malloc0 (sizeof *result);
 
@@ -603,7 +630,8 @@ lua_redis_callback_sync (redisAsyncContext *ac, gpointer r, gpointer priv)
 		 * This will call all callbacks pending so the entire context
 		 * will be destructed
 		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac, TRUE);
+		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+				RSPAMD_REDIS_RELEASE_FATAL);
 	}
 
 	result->result_ref = luaL_ref (L, LUA_REGISTRYINDEX);
@@ -630,13 +658,17 @@ lua_redis_callback_sync (redisAsyncContext *ac, gpointer r, gpointer priv)
 }
 
 static void
-lua_redis_timeout_sync (int fd, short what, gpointer priv)
+lua_redis_timeout_sync (EV_P_ ev_timer *w, int revents)
 {
-	struct lua_redis_request_specific_userdata *sp_ud = priv;
-	struct lua_redis_ctx *ctx = sp_ud->ctx;
+	struct lua_redis_request_specific_userdata *sp_ud =
+			(struct lua_redis_request_specific_userdata *)w->data;
+	struct lua_redis_ctx *ctx;
+	struct lua_redis_userdata *ud;
 	redisAsyncContext *ac;
 
-	msg_debug ("timeout while querying redis server: %p, redis: %p", sp_ud,
+	ud = sp_ud->c;
+	ctx = sp_ud->ctx;
+	msg_debug_lua_redis ("timeout while querying redis server: %p, redis: %p", sp_ud,
 			sp_ud->c->ctx);
 
 	if (sp_ud->c->ctx) {
@@ -652,14 +684,17 @@ lua_redis_timeout_sync (int fd, short what, gpointer priv)
 		 * This will call all callbacks pending so the entire context
 		 * will be destructed
 		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac, TRUE);
+		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+				RSPAMD_REDIS_RELEASE_FATAL);
 	}
 }
 
 static void
-lua_redis_timeout (int fd, short what, gpointer u)
+lua_redis_timeout (EV_P_ ev_timer *w, int revents)
 {
-	struct lua_redis_request_specific_userdata *sp_ud = u;
+	struct lua_redis_request_specific_userdata *sp_ud =
+			(struct lua_redis_request_specific_userdata *)w->data;
+	struct lua_redis_userdata *ud;
 	struct lua_redis_ctx *ctx;
 	redisAsyncContext *ac;
 
@@ -668,9 +703,10 @@ lua_redis_timeout (int fd, short what, gpointer u)
 	}
 
 	ctx = sp_ud->ctx;
+	ud = sp_ud->c;
 
 	REDIS_RETAIN (ctx);
-	msg_debug ("timeout while querying redis server: %p, redis: %p", sp_ud,
+	msg_debug_lua_redis ("timeout while querying redis server: %p, redis: %p", sp_ud,
 			sp_ud->c->ctx);
 	lua_redis_push_error ("timeout while connecting the server", ctx, sp_ud, TRUE);
 
@@ -684,7 +720,8 @@ lua_redis_timeout (int fd, short what, gpointer u)
 		 * This will call all callbacks pending so the entire context
 		 * will be destructed
 		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac, TRUE);
+		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+				RSPAMD_REDIS_RELEASE_FATAL);
 	}
 
 	REDIS_RELEASE (ctx);
@@ -790,17 +827,17 @@ lua_redis_parse_args (lua_State *L, gint idx, const gchar *cmd,
 static struct lua_redis_ctx *
 rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_async)
 {
-	struct lua_redis_ctx *ctx;
+	struct lua_redis_ctx *ctx = NULL;
 	rspamd_inet_addr_t *ip = NULL;
-	struct lua_redis_userdata *ud;
+	struct lua_redis_userdata *ud = NULL;
 	struct rspamd_lua_ip *addr = NULL;
 	struct rspamd_task *task = NULL;
-	const gchar *host;
-	const gchar *password = NULL, *dbname = NULL;
+	const gchar *host = NULL;
+	const gchar *password = NULL, *dbname = NULL, *log_tag = NULL;
 	gint cbref = -1;
 	struct rspamd_config *cfg = NULL;
 	struct rspamd_async_session *session = NULL;
-	struct event_base *ev_base = NULL;
+	struct ev_loop *ev_base = NULL;
 	gboolean ret = FALSE;
 	guint flags = 0;
 
@@ -850,8 +887,10 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 		else {
 			cfg = task->cfg;
 			session = task->s;
-			ev_base = task->ev_base;
+			ev_base = task->event_loop;
+			log_tag = task->task_pool->tag.uid;
 			ret = TRUE;
+
 		}
 
 		if (pcbref) {
@@ -873,11 +912,13 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 
 		if (lua_type (L, -1) == LUA_TUSERDATA) {
 			addr = lua_check_ip (L, -1);
+			host = rspamd_inet_address_to_string_pretty (addr->addr);
 		}
 		else if (lua_type (L, -1) == LUA_TSTRING) {
 			host = lua_tostring (L, -1);
 
-			if (rspamd_parse_inet_address (&ip, host, strlen (host))) {
+			if (rspamd_parse_inet_address (&ip,
+					host, strlen (host), RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 				addr = g_alloca (sizeof (*addr));
 				addr->addr = ip;
 
@@ -909,6 +950,13 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 		}
 		lua_pop (L, 1);
 
+		lua_pushstring (L, "no_pool");
+		lua_gettable (L, -2);
+		if (!!lua_toboolean (L, -1)) {
+			flags |= LUA_REDIS_NO_POOL;
+		}
+		lua_pop (L, 1);
+
 		lua_pop (L, 1); /* table */
 
 		if (session && rspamd_session_blocked (session)) {
@@ -933,9 +981,19 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 			ud->s = session;
 			ud->cfg = cfg;
 			ud->pool = cfg->redis_pool;
-			ud->ev_base = ev_base;
+			ud->event_loop = ev_base;
 			ud->task = task;
 
+			if (log_tag) {
+				rspamd_strlcpy (ud->log_tag, log_tag, sizeof (ud->log_tag));
+			}
+			else {
+				/* Use pointer itself as a tag */
+				rspamd_snprintf (ud->log_tag, sizeof (ud->log_tag),
+						"%ud",
+						(int)rspamd_cryptobox_fast_hash (&ud, sizeof (ud), 0));
+			}
+
 			if (task) {
 				ud->item = rspamd_symcache_get_cur_item (task);
 			}
@@ -967,11 +1025,13 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 			if (ud->ctx) {
 				msg_err_task_check ("cannot connect to redis: %s",
 						ud->ctx->errstr);
-				rspamd_redis_pool_release_connection (ud->pool, ud->ctx, TRUE);
+				rspamd_redis_pool_release_connection (ud->pool, ud->ctx,
+						RSPAMD_REDIS_RELEASE_FATAL);
 				ud->ctx = NULL;
 			}
 			else {
-				msg_err_task_check ("cannot connect to redis: unknown error");
+				msg_err_task_check ("cannot connect to redis (OS error): %s",
+						strerror (errno));
 			}
 
 			REDIS_RELEASE (ctx);
@@ -979,6 +1039,8 @@ rspamd_lua_redis_prepare_connection (lua_State *L, gint *pcbref, gboolean is_asy
 			return NULL;
 		}
 
+		msg_debug_lua_redis ("opened redis connection host=%s; ctx=%p; ud=%p",
+				host, ctx, ud);
 
 		return ctx;
 	}
@@ -1009,7 +1071,6 @@ lua_redis_make_request (lua_State *L)
 	struct lua_redis_userdata *ud;
 	struct lua_redis_ctx *ctx, **pctx;
 	const gchar *cmd = NULL;
-	struct timeval tv;
 	gdouble timeout = REDIS_DEFAULT_TIMEOUT;
 	gint cbref = -1;
 	gboolean ret = FALSE;
@@ -1064,15 +1125,16 @@ lua_redis_make_request (lua_State *L)
 
 			REDIS_RETAIN (ctx); /* Cleared by fin event */
 			ctx->cmds_pending ++;
-			double_to_tv (timeout, &tv);
-			event_set (&sp_ud->timeout, -1, EV_TIMEOUT, lua_redis_timeout, sp_ud);
-			event_base_set (ud->ev_base, &sp_ud->timeout);
-			event_add (&sp_ud->timeout, &tv);
+			sp_ud->timeout_ev.data = sp_ud;
+			ev_now_update_if_cheap ((struct ev_loop *)ud->event_loop);
+			ev_timer_init (&sp_ud->timeout_ev, lua_redis_timeout, timeout, 0.0);
+			ev_timer_start (ud->event_loop, &sp_ud->timeout_ev);
 			ret = TRUE;
 		}
 		else {
 			msg_info ("call to redis failed: %s", ud->ctx->errstr);
-			rspamd_redis_pool_release_connection (ud->pool, ud->ctx, TRUE);
+			rspamd_redis_pool_release_connection (ud->pool, ud->ctx,
+					RSPAMD_REDIS_RELEASE_FATAL);
 			ud->ctx = NULL;
 			REDIS_RELEASE (ctx);
 			ret = FALSE;
@@ -1139,7 +1201,8 @@ lua_redis_make_request_sync (lua_State *L)
 		}
 		else if (lua_type (L, -1) == LUA_TSTRING) {
 			host = lua_tostring (L, -1);
-			if (rspamd_parse_inet_address (&ip, host, strlen (host))) {
+			if (rspamd_parse_inet_address (&ip,
+					host, strlen (host), RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
 				addr = g_alloca (sizeof (*addr));
 				addr->addr = ip;
 
@@ -1347,7 +1410,6 @@ lua_redis_add_cmd (lua_State *L)
 	const gchar *cmd = NULL;
 	gint args_pos = 2;
 	gint cbref = -1, ret;
-	struct timeval tv;
 
 	if (ctx) {
 		if (ctx->flags & LUA_REDIS_TERMINATED) {
@@ -1426,19 +1488,18 @@ lua_redis_add_cmd (lua_State *L)
 				}
 			}
 
-			double_to_tv (sp_ud->c->timeout, &tv);
+			sp_ud->timeout_ev.data = sp_ud;
 
 			if (IS_ASYNC (ctx)) {
-				event_set (&sp_ud->timeout, -1, EV_TIMEOUT,
-						lua_redis_timeout, sp_ud);
+				ev_timer_init (&sp_ud->timeout_ev, lua_redis_timeout,
+						sp_ud->c->timeout, 0.0);
 			}
 			else {
-				event_set (&sp_ud->timeout, -1, EV_TIMEOUT,
-						lua_redis_timeout_sync, sp_ud);
+				ev_timer_init (&sp_ud->timeout_ev, lua_redis_timeout_sync,
+						sp_ud->c->timeout, 0.0);
 			}
 
-			event_base_set (ud->ev_base, &sp_ud->timeout);
-			event_add (&sp_ud->timeout, &tv);
+			ev_timer_start (ud->event_loop, &sp_ud->timeout_ev);
 			REDIS_RETAIN (ctx);
 			ctx->cmds_pending ++;
 		}
@@ -1579,17 +1640,7 @@ lua_load_redis (lua_State * L)
 void
 luaopen_redis (lua_State * L)
 {
-	luaL_newmetatable (L, "rspamd{redis}");
-	lua_pushstring (L, "__index");
-	lua_pushvalue (L, -2);
-	lua_settable (L, -3);
-
-	lua_pushstring (L, "class");
-	lua_pushstring (L, "rspamd{redis}");
-	lua_rawset (L, -3);
-
-	luaL_register (L, NULL, redislib_m);
+	rspamd_lua_new_class (L, "rspamd{redis}", redislib_m);
 	lua_pop (L, 1);
-
 	rspamd_lua_add_preload (L, "rspamd_redis", lua_load_redis);
 }
