@@ -16,15 +16,17 @@
 
 #include "config.h"
 #include "rspamadm.h"
-#include "libutil/http_connection.h"
-#include "libutil/http_private.h"
-#include "libutil/http_router.h"
+#include "libserver/http/http_connection.h"
+#include "libserver/http/http_private.h"
+#include "libserver/http/http_router.h"
 #include "printf.h"
 #include "lua/lua_common.h"
 #include "lua/lua_thread_pool.h"
 #include "message.h"
 #include "unix-std.h"
-#include "linenoise.h"
+#ifdef WITH_LUA_REPL
+#include "replxx.h"
+#endif
 #include "worker_util.h"
 #ifdef WITH_LUAJIT
 #include <luajit.h>
@@ -38,11 +40,14 @@ static guint max_history = 2000;
 static gchar *serve = NULL;
 static gchar *exec_line = NULL;
 static gint batch = -1;
-static gboolean per_line = FALSE;
 extern struct rspamd_async_session *rspamadm_session;
 
 static const char *default_history_file = ".rspamd_repl.hist";
 
+#ifdef WITH_LUA_REPL
+static Replxx *rx_instance = NULL;
+#endif
+
 #ifdef WITH_LUAJIT
 #define MAIN_PROMPT LUAJIT_VERSION "> "
 #else
@@ -119,8 +124,6 @@ static GOptionEntry entries[] = {
 				"Serve http lua server", NULL},
 		{"batch", 'b', 0, G_OPTION_ARG_NONE, &batch,
 				"Batch execution mode", NULL},
-		{"per-line", 'p', 0, G_OPTION_ARG_NONE, &per_line,
-				"Pass each line of input to the specified lua script", NULL},
 		{"exec", 'e', 0, G_OPTION_ARG_STRING, &exec_line,
 				"Execute specified script", NULL},
 		{"args", 'a', 0, G_OPTION_ARG_STRING_ARRAY, &lua_args,
@@ -222,15 +225,12 @@ rspamadm_lua_load_script (lua_State *L, const gchar *path)
 		return FALSE;
 	}
 
-	if (!per_line) {
-
-		if (lua_repl_thread_call (thread, 0, (void *)path, lua_thread_str_error_cb) != 0) {
-			return FALSE;
-		}
-
-		lua_settop (L, 0);
+	if (lua_repl_thread_call (thread, 0, (void *)path, lua_thread_str_error_cb) != 0) {
+		return FALSE;
 	}
 
+	lua_settop (L, 0);
+
 	return TRUE;
 }
 
@@ -268,24 +268,22 @@ rspamadm_exec_input (lua_State *L, const gchar *input)
 
 	g_string_free (tb, TRUE);
 
-	if (!per_line) {
 
-		top = lua_gettop (L);
+	top = lua_gettop (L);
 
-		if (lua_repl_thread_call (thread, 0, NULL, NULL) == 0) {
-			/* Print output */
-			for (i = top; i <= lua_gettop (L); i++) {
-				if (lua_isfunction (L, i)) {
-					lua_pushvalue (L, i);
-					cbref = luaL_ref (L, LUA_REGISTRYINDEX);
+	if (lua_repl_thread_call (thread, 0, NULL, NULL) == 0) {
+		/* Print output */
+		for (i = top; i <= lua_gettop (L); i++) {
+			if (lua_isfunction (L, i)) {
+				lua_pushvalue (L, i);
+				cbref = luaL_ref (L, LUA_REGISTRYINDEX);
 
-					rspamd_printf ("local function: %d\n", cbref);
-				} else {
-					memset (&tr, 0, sizeof (tr));
-					lua_logger_out_type (L, i, outbuf, sizeof (outbuf), &tr,
-							LUA_ESCAPE_UNPRINTABLE);
-					rspamd_printf ("%s\n", outbuf);
-				}
+				rspamd_printf ("local function: %d\n", cbref);
+			} else {
+				memset (&tr, 0, sizeof (tr));
+				lua_logger_out_type (L, i, outbuf, sizeof (outbuf) - 1, &tr,
+						LUA_ESCAPE_UNPRINTABLE);
+				rspamd_printf ("%s\n", outbuf);
 			}
 		}
 	}
@@ -296,7 +294,7 @@ wait_session_events (void)
 {
 	/* XXX: it's probably worth to add timeout here - not to wait forever */
 	while (rspamd_session_events_pending (rspamadm_session) > 0) {
-		event_base_loop (rspamd_main->ev_base, EVLOOP_ONCE);
+		ev_loop (rspamd_main->event_loop, EVRUN_ONCE);
 	}
 }
 
@@ -433,7 +431,7 @@ rspamadm_lua_message_handler (lua_State *L, gint argc, gchar **argv)
 			rspamd_printf ("cannot open %s: %s\n", argv[i], strerror (errno));
 		}
 		else {
-			task = rspamd_task_new (NULL, rspamd_main->cfg, NULL, NULL, NULL);
+			task = rspamd_task_new (NULL, rspamd_main->cfg, NULL, NULL, NULL, FALSE);
 
 			if (!rspamd_task_load_message (task, NULL, map, len)) {
 				rspamd_printf ("cannot load %s\n", argv[i]);
@@ -510,17 +508,143 @@ rspamadm_lua_try_dot_command (lua_State *L, const gchar *input)
 	return FALSE;
 }
 
+#ifdef WITH_LUA_REPL
+static gint lex_ref_idx = -1;
+
 static void
-rspamadm_lua_run_repl (lua_State *L)
+lua_syntax_highlighter (const char *str, ReplxxColor *colours, int size, void *ud)
+{
+	lua_State *L = (lua_State *)ud;
+
+	if (lex_ref_idx == -1) {
+		if (!rspamd_lua_require_function (L, "lua_lexer", "lex_to_table")) {
+			fprintf (stderr, "cannot require lua_lexer!\n");
+
+			exit (EXIT_FAILURE);
+		}
+
+		lex_ref_idx = luaL_ref (L, LUA_REGISTRYINDEX);
+	}
+
+	lua_rawgeti (L, LUA_REGISTRYINDEX, lex_ref_idx);
+	lua_pushstring (L, str);
+
+	if (lua_pcall (L, 1, 1, 0) != 0) {
+		fprintf (stderr, "cannot lex a string!\n");
+	}
+	else {
+		/* Process what we have after lexing */
+		gsize nelts = rspamd_lua_table_size (L, -1);
+
+		for (gsize i = 0; i < nelts; i ++) {
+			/*
+			 * Indexes in the table:
+			 * 1 - type of element (string)
+			 * 2 - text (string)
+			 * 3 - line num (int), always 1...
+			 * 4 - column num (must be less than size)
+			 */
+			const gchar *what;
+			gsize column, tlen, cur_top, elt_pos;
+			ReplxxColor elt_color = REPLXX_COLOR_DEFAULT;
+
+			cur_top = lua_gettop (L);
+			lua_rawgeti (L, -1, i + 1);
+			elt_pos = lua_gettop (L);
+			lua_rawgeti (L, elt_pos, 1);
+			what = lua_tostring (L, -1);
+			lua_rawgeti (L, elt_pos, 2);
+			lua_tolstring (L, -1, &tlen);
+			lua_rawgeti (L, elt_pos, 4);
+			column = lua_tointeger (L, -1);
+
+			g_assert (column > 0);
+			column --; /* Start from 0 */
+
+			if (column + tlen > size) {
+				/* Likely utf8 case, too complicated to match */
+				lua_settop (L, cur_top);
+				continue;
+			}
+
+			/* Check what and adjust color */
+			if (strcmp (what, "identifier") == 0) {
+				elt_color = REPLXX_COLOR_NORMAL;
+			}
+			else if (strcmp (what, "number") == 0) {
+				elt_color = REPLXX_COLOR_BLUE;
+			}
+			else if (strcmp (what, "string") == 0) {
+				elt_color = REPLXX_COLOR_GREEN;
+			}
+			else if (strcmp (what, "keyword") == 0) {
+				elt_color = REPLXX_COLOR_WHITE;
+			}
+			else if (strcmp (what, "constant") == 0) {
+				elt_color = REPLXX_COLOR_WHITE;
+			}
+			else if (strcmp (what, "operator") == 0) {
+				elt_color = REPLXX_COLOR_CYAN;
+			}
+			else if (strcmp (what, "comment") == 0) {
+				elt_color = REPLXX_COLOR_BRIGHTGREEN;
+			}
+			else if (strcmp (what, "error") == 0) {
+				elt_color = REPLXX_COLOR_ERROR;
+			}
+
+			for (gsize j = column; j < column + tlen; j ++) {
+				colours[j] = elt_color;
+			}
+
+			/* Restore stack */
+			lua_settop (L, cur_top);
+		}
+	}
+
+	lua_settop (L, 0);
+}
+#endif
+
+static void
+rspamadm_lua_run_repl (lua_State *L, bool is_batch)
 {
 	gchar *input;
+#ifdef WITH_LUA_REPL
 	gboolean is_multiline = FALSE;
-	GString *tb;
-	guint i;
+	GString *tb = NULL;
+	gsize i;
+#endif
 
 	for (;;) {
+#ifndef WITH_LUA_REPL
+		size_t linecap = 0;
+		ssize_t linelen;
+
+		fprintf (stdout, "%s ", MAIN_PROMPT);
+
+		linelen = getline (&input, &linecap, stdin);
+
+		if (linelen > 0) {
+			if (input[linelen - 1] == '\n') {
+				linelen --;
+			}
+
+			rspamadm_exec_input (L, input);
+		}
+		else {
+			break;
+		}
+
+		lua_settop (L, 0);
+#else
+		if (!is_batch) {
+			replxx_set_highlighter_callback (rx_instance, lua_syntax_highlighter,
+					L);
+		}
+
 		if (!is_multiline) {
-			input = linenoise (MAIN_PROMPT);
+			input = (gchar *)replxx_input (rx_instance, MAIN_PROMPT);
 
 			if (input == NULL) {
 				return;
@@ -528,26 +652,27 @@ rspamadm_lua_run_repl (lua_State *L)
 
 			if (input[0] == '.') {
 				if (rspamadm_lua_try_dot_command (L, input)) {
-					linenoiseHistoryAdd (input);
-					linenoiseFree (input);
+					if (!is_batch) {
+						replxx_history_add (rx_instance, input);
+					}
 					continue;
 				}
 			}
 
 			if (strcmp (input, "{{") == 0) {
 				is_multiline = TRUE;
-				linenoiseFree (input);
 				tb = g_string_sized_new (8192);
 				continue;
 			}
 
 			rspamadm_exec_input (L, input);
-			linenoiseHistoryAdd (input);
-			linenoiseFree (input);
+			if (!is_batch) {
+				replxx_history_add (rx_instance, input);
+			}
 			lua_settop (L, 0);
 		}
 		else {
-			input = linenoise (MULTILINE_PROMPT);
+			input = (gchar *)replxx_input (rx_instance, MULTILINE_PROMPT);
 
 			if (input == NULL) {
 				g_string_free (tb, TRUE);
@@ -556,7 +681,6 @@ rspamadm_lua_run_repl (lua_State *L)
 
 			if (strcmp (input, "}}") == 0) {
 				is_multiline = FALSE;
-				linenoiseFree (input);
 				rspamadm_exec_input (L, tb->str);
 
 				/* Replace \n with ' ' for sanity */
@@ -566,15 +690,17 @@ rspamadm_lua_run_repl (lua_State *L)
 					}
 				}
 
-				linenoiseHistoryAdd (tb->str);
+				if (!is_batch) {
+					replxx_history_add (rx_instance, tb->str);
+				}
 				g_string_free (tb, TRUE);
 			}
 			else {
 				g_string_append (tb, input);
 				g_string_append (tb, " \n");
-				linenoiseFree (input);
 			}
 		}
+#endif
 	}
 }
 
@@ -591,15 +717,16 @@ struct rspamadm_lua_repl_session {
 };
 
 static void
-rspamadm_lua_accept_cb (gint fd, short what, void *arg)
+rspamadm_lua_accept_cb (EV_P_ ev_io *w, int revents)
 {
-	struct rspamadm_lua_repl_context *ctx = arg;
+	struct rspamadm_lua_repl_context *ctx =
+			(struct rspamadm_lua_repl_context *)w->data;
 	rspamd_inet_addr_t *addr;
 	struct rspamadm_lua_repl_session *session;
 	gint nfd;
 
 	if ((nfd =
-			rspamd_accept_from_socket (fd, &addr, NULL)) == -1) {
+			rspamd_accept_from_socket (w->fd, &addr, NULL, NULL)) == -1) {
 		rspamd_fprintf (stderr, "accept failed: %s", strerror (errno));
 		return;
 	}
@@ -657,15 +784,11 @@ rspamadm_lua_handle_exec (struct rspamd_http_connection_entry *conn_ent,
 	GString *tb;
 	gint err_idx, i;
 	lua_State *L;
-	struct rspamadm_lua_repl_context *ctx;
-	struct rspamadm_lua_repl_session *session = conn_ent->ud;
 	ucl_object_t *obj, *elt;
 	const gchar *body;
 	gsize body_len;
-
-	ctx = session->ctx;
-
 	struct thread_entry *thread = lua_thread_pool_get_for_config (rspamd_main->cfg);
+
 	L = thread->lua_state;
 
 	body = rspamd_http_message_get_body (msg, &body_len);
@@ -744,9 +867,12 @@ rspamadm_lua (gint argc, gchar **argv, const struct rspamadm_command *cmd)
 	if (!g_option_context_parse (context, &argc, &argv, &error)) {
 		fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_error_free (error);
+		g_option_context_free (context);
 		exit (1);
 	}
 
+	g_option_context_free (context);
+
 	if (batch == -1) {
 		if (isatty (STDIN_FILENO)) {
 			batch = 0;
@@ -793,22 +919,22 @@ rspamadm_lua (gint argc, gchar **argv, const struct rspamadm_command *cmd)
 		/* HTTP Server mode */
 		GPtrArray *addrs = NULL;
 		gchar *name = NULL;
-		struct event_base *ev_base;
+		struct ev_loop *ev_base;
 		struct rspamd_http_connection_router *http;
 		gint fd;
 		struct rspamadm_lua_repl_context *ctx;
 
-		if (!rspamd_parse_host_port_priority (serve, &addrs, NULL, &name,
-				10000, NULL)) {
+		if (rspamd_parse_host_port_priority (serve, &addrs, NULL, &name,
+				10000, TRUE, NULL) == RSPAMD_PARSE_ADDR_FAIL) {
 			fprintf (stderr, "cannot listen on %s", serve);
 			exit (EXIT_FAILURE);
 		}
 
-		ev_base = rspamd_main->ev_base;
+		ev_base = rspamd_main->event_loop;
 		ctx = g_malloc0  (sizeof (*ctx));
 		http = rspamd_http_router_new (rspamadm_lua_error_handler,
 						rspamadm_lua_finish_handler,
-						NULL,
+						0.0,
 						NULL,
 						rspamd_main->http_ctx);
 		ctx->L = L;
@@ -822,19 +948,17 @@ rspamadm_lua (gint argc, gchar **argv, const struct rspamadm_command *cmd)
 
 			fd = rspamd_inet_address_listen (addr, SOCK_STREAM, TRUE);
 			if (fd != -1) {
-				struct event *ev;
+				static ev_io ev;
 
-				ev = g_malloc0 (sizeof (*ev));
-				event_set (ev, fd, EV_READ|EV_PERSIST, rspamadm_lua_accept_cb,
-						ctx);
-				event_base_set (ev_base, ev);
-				event_add (ev, NULL);
+				ev.data = ctx;
+				ev_io_init (&ev, rspamadm_lua_accept_cb, fd, EV_READ);
+				ev_io_start (ev_base, &ev);
 				rspamd_printf ("listen on %s\n",
 						rspamd_inet_address_to_string_pretty (addr));
 			}
 		}
 
-		event_base_loop (ev_base, 0);
+		ev_loop (ev_base, 0);
 
 		exit (EXIT_SUCCESS);
 	}
@@ -877,63 +1001,23 @@ rspamadm_lua (gint argc, gchar **argv, const struct rspamadm_command *cmd)
 		g_hash_table_insert (cmds_hash, (gpointer)cmds[i].name, &cmds[i]);
 	}
 
-	if (per_line) {
-		GIOChannel *in;
-		GString *buf;
-		gsize end_pos;
-		GIOStatus ret;
-		gint old_top;
-		GError *err = NULL;
 
-		in = g_io_channel_unix_new (STDIN_FILENO);
-		buf = g_string_sized_new (BUFSIZ);
-
-again:
-		while ((ret = g_io_channel_read_line_string (in, buf, &end_pos, &err)) ==
-				G_IO_STATUS_NORMAL) {
-			old_top = lua_gettop (L);
-			lua_pushvalue (L, -1);
-			lua_pushlstring (L, buf->str, MIN (buf->len, end_pos));
-			lua_setglobal (L, "input");
-
-			struct thread_entry *thread = lua_thread_pool_get_for_config (rspamd_main->cfg);
-			L = thread->lua_state;
-
-			lua_repl_thread_call (thread, 0, NULL, NULL);
-
-			lua_settop (L, old_top);
-		}
-
-		if (ret == G_IO_STATUS_AGAIN) {
-			goto again;
-		}
-
-		g_string_free (buf, TRUE);
-		g_io_channel_shutdown (in, FALSE, NULL);
-
-		if (ret == G_IO_STATUS_EOF) {
-			if (err) {
-				g_error_free (err);
-			}
-		}
-		else {
-			rspamd_fprintf (stderr, "IO error: %e\n", err);
-
-			if (err) {
-				g_error_free (err);
-			}
-
-			exit (-errno);
-		}
-	}
-	else {
-		if (!batch) {
-			linenoiseHistorySetMaxLen (max_history);
-			linenoiseHistoryLoad (histfile);
-			rspamadm_lua_run_repl (L);
-			linenoiseHistorySave (histfile);
-		} else {
-			rspamadm_lua_run_repl (L);
-		}
+#ifdef WITH_LUA_REPL
+	rx_instance = replxx_init ();
+#endif
+	if (!batch) {
+#ifdef WITH_LUA_REPL
+		replxx_set_max_history_size (rx_instance, max_history);
+		replxx_history_load (rx_instance, histfile);
+#endif
+		rspamadm_lua_run_repl (L, false);
+#ifdef WITH_LUA_REPL
+		replxx_history_save (rx_instance, histfile);
+#endif
+	} else {
+		rspamadm_lua_run_repl (L, true);
 	}
+#ifdef WITH_LUA_REPL
+	replxx_end (rx_instance);
+#endif
 }
